from typing import Optional
import warnings
import itertools
import operator
import copy

import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from pandas.core.generic import NDFrame # type:ignore

import matplotlib.pyplot as plt  # type: ignore
import seaborn as sns  # type: ignore
sns.set(style="white", color_codes=True)

# imported ARIMA from statsmodels pkg
from statsmodels.tsa.arima_model import ARIMA # type: ignore

# helper functions
from ...utils import print_static_rmse, print_dynamic_rmse
from ...models.ar_based.param_finder import find_lowest_pq
import pdb


class BuildArima():
    def __init__(self, metric='aic', p_max=3, d_max=1, q_max=3, forecast_period=2, method='mle', verbose=0):
        """
        Automatically build an ARIMA Model
        """
        self.metric = metric
        self.p_max = p_max
        self.d_max = d_max
        self.q_max = q_max
        self.forecast_period = forecast_period
        self.method = method
        self.verbose = verbose
        self.model = None

    def fit(self, ts_df):
        """
        Build a Time Series Model using SARIMAX from statsmodels.

        This builds a Non Seasonal ARIMA model given a Univariate time series dataframe with time
        as the Index, ts_df can be a dataframe with one column only or a single array. Dont send
        Multiple Columns!!! Include only that variable that is a Time Series. DO NOT include
        Non-Stationary data. Make sure your Time Series is "Stationary"!! If not, this
        will give spurious results, since it automatically builds a Non-Seasonal model,
        you need not give it a Seasonal True/False flag.
        "metric": You can give it any of the following metrics as criteria: AIC, BIC, Deviance,
        Log-likelihood. Optionally, you can give it a fit method as one of the following:
        {'css-mle','mle','css'}
        """

        solver = 'lbfgs'  # default

        p_min = 0
        d_min = 0
        q_min = 0
        # Initialize a DataFrame to store the results
        iteration = 0
        results_dict = {}

        ################################################################################
        ####### YOU MUST Absolutely set this parameter correctly as "levels". If not,
        ####  YOU WILL GET DIFFERENCED PREDICTIONS WHICH ARE FIENDISHLY DIFFICULT TO UNDO.
        #### If you set this to levels, then you can do any order of differencing and
        ####  ARIMA will give you predictions in the same level as orignal values.
        ################################################################################
        pred_type = 'levels'
        #########################################################################
        ts_train = ts_df[:-self.forecast_period]
        ts_test = ts_df[-self.forecast_period:]
        if self.verbose == 1:
            print('Data Set split into train %s and test %s for Cross Validation Purposes'
                % (ts_train.shape, ts_test.shape))
        #########################################################################
        if ts_train.dtype == 'int64':
            ts_train = ts_train.astype(float)
        for d_val in range(d_min, self.d_max+1):
            print('\nDifferencing = %d' % d_val)
            results_bic = pd.DataFrame(
                index=['AR{}'.format(i) for i in range(p_min, self.p_max+1)],
                columns=['MA{}'.format(i) for i in range(q_min, self.q_max+1)]
            )
            for p_val, q_val in itertools.product(range(p_min, self.p_max+1), range(q_min, self.q_max+1)):
                if p_val == 0 and d_val == 0 and q_val == 0:
                    results_bic.loc['AR{}'.format(p_val), 'MA{}'.format(q_val)] = np.nan
                    continue
                else:
                    try:
                        model = ARIMA(ts_train, order=(p_val, d_val, q_val))
                        results = model.fit(transparams=False, method=self.method, solver=solver, disp=False)
                        results_bic.loc['AR{}'.format(p_val), 'MA{}'.format(q_val)] = eval('results.' + self.metric)
                        if iteration % 10 == 0:
                            print(' Iteration %d completed...' % iteration)
                        iteration += 1
                        if iteration >= 100:
                            print('    Ending Iterations at %d' % iteration)
                            break
                    except:
                        iteration += 1
                        continue
            results_bic = results_bic[results_bic.columns].astype(float)
            interim_d = copy.deepcopy(d_val)
            interim_p, interim_q, interim_bic = find_lowest_pq(results_bic)
            if self.verbose == 1:
                _, ax = plt.subplots(figsize=(20, 10))
                ax = sns.heatmap(results_bic,
                                mask=results_bic.isnull(),
                                ax=ax,
                                annot=True,
                                fmt='.0f')
                ax.set_title(self.metric)
            results_dict[str(interim_p) + ' ' + str(interim_d) + ' ' + str(interim_q)] = interim_bic
        best_bic = min(results_dict.items(), key=operator.itemgetter(1))[1]
        best_pdq = min(results_dict.items(), key=operator.itemgetter(1))[0]
        best_p = int(best_pdq.split(' ')[0])
        best_d = int(best_pdq.split(' ')[1])
        best_q = int(best_pdq.split(' ')[2])
        print('\nBest model is: Non Seasonal ARIMA(%d,%d,%d), %s = %0.3f' % (best_p, best_d, best_q, self.metric, best_bic))
        bestmodel = ARIMA(ts_train, order=(best_p, best_d, best_q))
        print('####    Fitting best model for full data set now. Will take time... ######')
        try:
            self.model = bestmodel.fit(transparams=True, method=self.method, solver=solver, disp=False)
        except:
            self.model = bestmodel.fit(transparams=False, method=self.method, solver=solver, disp=False)
        ### this is needed for static forecasts ####################
        y_truth = ts_train[:]
        y_forecasted = self.model.predict(typ='levels')
        concatenated = pd.concat([y_truth, y_forecasted], axis=1, keys=['original', 'predicted'])
        if best_d == 0:
            #### Do this for ARIMA only ######
            ###  If there is no differencing DO NOT use predict_type since it will give an error = do not use "linear".
            print('Static Forecasts:')
            print_static_rmse(concatenated['original'].values, concatenated['predicted'].values, best_d)
            start_date = ts_df.index[-self.forecast_period]
            end_date = ts_df.index[-1]
            pred_dynamic = self.model.predict(start=start_date, end=end_date, dynamic=True)
            if self.verbose == 1:
                ax = concatenated[['original', 'predicted']][best_d:].plot()
                pred_dynamic.plot(label='Dynamic Forecast', ax=ax, figsize=(15, 5))
                print('Dynamic %d-period Forecasts:' % (self.forecast_period,))
                plt.legend()
                plt.show(block=False)
        else:
            #### Do this for ARIMA only ######
            ####  If there is differencing, you must use "levels" as the predict type to get original levels as actuals
            pred_type = 'levels'
            print('Static Forecasts:')
            print_static_rmse(y_truth[best_d:], y_forecasted)
            ########### Dynamic One Step Ahead Forecast ###########################
            ### Dynamic Forecasts are a better representation of true predictive power
            ## since they only use information from the time series up to a certain point,
            ## and after that, forecasts are generated using values from previous forecasted
            ## time points.
            #################################################################################

            # TODO: Check if this can be changed to use predict function directly.
            start_date = ts_df.index[-self.forecast_period]
            end_date = ts_df.index[-1]
            pred_dynamic = self.model.predict(typ=pred_type, start=start_date, end=end_date, dynamic=True)
            try:
                pred_dynamic[pd.to_datetime((pred_dynamic.index-best_d).values[0])] = \
                                        y_truth[pd.to_datetime((pred_dynamic.index-best_d).values[0])]
            except:
                print('Dynamic predictions erroring but continuing...')
            pred_dynamic.sort_index(inplace=True)
            print('\nDynamic %d-period Forecasts:' % self.forecast_period)
            if self.verbose == 1:
                ax = concatenated.plot()
                pred_dynamic.plot(label='Dynamic Forecast', ax=ax, figsize=(15, 5))
                ax.set_xlabel('Date')
                ax.set_ylabel('Values')
                plt.legend()
                plt.show(block=False)
        if self.verbose == 1:
            try:
                self.model.plot_diagnostics(figsize=(16, 12))
            except:
                pass
        print(self.model.summary())

        res_frame = self.predict(simple=False)

        if self.verbose == 1:
            print('Model Forecast(s):\n', res_frame)
        rmse, norm_rmse = print_dynamic_rmse(ts_test, pred_dynamic, ts_train)
        return self.model, res_frame, rmse, norm_rmse

    def predict(
        self,
        testdata: Optional[pd.DataFrame]=None,
        forecast_period: Optional[int] = None,
        simple: bool = True) -> NDFrame:
        """
        Return the predictions
        # TODO: Check if the series can be converted to a dataframe for all models.
        :rtype cam be Pandas Series (simple), pandas dataframe (simple = False) or None
        """

        # TODO: Add processing of 'simple' argument and return type

        if testdata is not None:
            warnings.warn(
                "You have passed exogenous variables to make predictions for a ARIMA model." +
                "ARIMA models are univariate models and hence these exogenous variables will be ignored for these predictions."
            )

        # TODO: Predictions coming from ARIMA include extra information compared to SARIMAX and VAR.
        # Need to make it consistent
        # Extract the dynamic predicted and true values of our time series
        if forecast_period is None:
            # use the forecast period used during training
            forecast_period = self.forecast_period

        y_forecasted = self.model.forecast(forecast_period)


        # TODO: Check if the datetime index can be obtained as in the case of SARIMAX.
        # Currently it is just a text index, e.g. Forecast_1, ...
        if simple:
            res_frame = pd.DataFrame([
                y_forecasted[0], # Mean Forecast
                ],
                index=['mean'],
                columns=['Forecast_' + str(x) for x in range(1, forecast_period+1)]
            ).T
            res_frame = res_frame.squeeze() # Convert to a pandas series object
        else:
            res_frame = pd.DataFrame([
                y_forecasted[0], # Mean Forecast
                y_forecasted[1], # Std Error
                y_forecasted[2], # Lower and Upper CI
                ],
                index=['mean','mean_se','mean_ci'],
                columns=['Forecast_' + str(x) for x in range(1, forecast_period+1)]
            ).T

            res_frame['mean_ci_lower'] = res_frame['mean_ci'].map(lambda x: x[0])
            res_frame['mean_ci_upper'] = res_frame['mean_ci'].map(lambda x: x[1])
            res_frame.drop('mean_ci', axis=1, inplace=True)

        return res_frame



from typing import List
import numpy as np
import pandas as pd  # type: ignore
import copy
import pdb
from sklearn.model_selection import TimeSeriesSplit  # type: ignore
import dask
import dask.dataframe as dd
from pandas.api.types import is_datetime64_any_dtype as is_datetime

##### This function loads a time series data and sets the index as a time series
def load_ts_data(filename, ts_column, sep, target, dask_xgboost_flag=0):
    """
    This function loads a given filename into a pandas dataframe and sets the
    ts_column as a Time Series index. Note that filename should contain the full
    path to the file.
    Inputs:
        filename: name of file that contains data
        ts_column: name of time series column in data
        sep: separator used as a column separator in datafile
        target: name of the target column to predict
        dask_xgboost_flag: flag that will tell whether to load into dask or pandas dataframe.        
        If dask_xgboost_flag is set to True it returns both a dask as well as pandas DataFrame.
        If dask_xgboost_flag is set to False it returns both of them as pandas DataFrames.

    Outputs:
        dft: dask DataFrame
        filename: pandas DataFrame
    """
    
    if isinstance(filename, str):
        filename = pd.read_csv(filename, sep=sep, index_col=ts_column, parse_dates=True)
    ### If filename is not a string, it must be a dataframe and can be loaded
    if dask_xgboost_flag:
        if type(filename) == dask.dataframe.core.DataFrame:
            print('    Since dask_xgboost_flag is True, and input is dask, continuing...')
        else:
            filename = copy.deepcopy(filename)
            print('    Since dask_xgboost_flag is True and input is pandas, reducing memory size of df and loading into dask')
            filename = reduce_mem_usage(filename)
            dft =   dd.from_pandas(filename, npartitions=1)
            print('    Converted pandas dataframe into a Dask dataframe ...' )
    else:
        dft = copy.deepcopy(filename)
        print('    Using given input: pandas dataframe...')
    ##################    L O A D    T E S T   D A T A      ######################
    dft = remove_duplicate_cols_in_dataset(dft)
    #######   Make sure you change it to a date-time index #####
    if dask_xgboost_flag:
        ### if dask exists, you need to change its datetime index also ##
        dft, _ = change_to_datetime_index(dft, ts_column)
    ### you have to change the pandas df also to datetime index ###
    filename, _ = change_to_datetime_index(filename, ts_column)    
    #preds = [x for x in list(dft) if x not in [target]]
    #dft = dft[[target]+preds]
    return dft, filename
####################################################################################################################
def load_test_data(filename, ts_column, sep, target, dask_xgboost_flag=0):
    """
    This function loads a given filename into a pandas dataframe and sets the
    ts_column as a Time Series index. Note that filename should contain the full
    path to the file.
    """
    
    if isinstance(filename, str):
        filename = pd.read_csv(filename, sep=sep, index_col=ts_column, parse_dates=True)
        ### If filename is not a string, it must be a dataframe and can be loaded
    else:
        if type(filename) == dask.dataframe.core.DataFrame:
            print('    Since dask_xgboost_flag is True, and input is dask, continuing...')
            ddf =   filename.compute()
            print('    Converted dask dataframe into a pandas dataframe ...' )
            print('    Reducing memory size of df and loading into dask')
            dft = reduce_mem_usage(ddf)
        else:
            dft = copy.deepcopy(filename)
            print('    Using given input: pandas dataframe...')
    ##################    L O A D    T E S T   D A T A      ######################
    dft = remove_duplicate_cols_in_dataset(dft)
    return dft
####################################################################################################################
def remove_duplicate_cols_in_dataset(df):
    df = copy.deepcopy(df)
    cols = df.columns.tolist()
    number_duplicates = df.columns.duplicated().astype(int).sum()
    if  number_duplicates > 0:
        print('Detected %d duplicate columns in dataset. Removing duplicates...' %number_duplicates)
        df = df.loc[:,~df.columns.duplicated()]
    return df
###########################################################################
def change_to_datetime_index(dft, ts_column):
    dft = copy.deepcopy(dft)
    if isinstance(dft, pd.Series) or isinstance(dft, pd.DataFrame):
        try:
            ### If ts_column is not a string column, then set its format to an empty string ##
            str_format = ''
            ############### Check if it has an index or a column with the name of train time series column ####
            
            if ts_column in dft.columns:
                print('    %s column exists in given train data...' %ts_column)
                str_first_value = dft[ts_column].values[0]
                str_values = dft[ts_column].values[:12] ### we want to test a big sample of them 
                if type(str_first_value) == str:
                    ### if it is an object column, convert ts_column into datetime and then set as index
                    str_format = infer_date_time_format(str_values)
                    if str_format:
                        str_format = str_format[0]
                        ts_index = pd.to_datetime(dft.pop(ts_column), format=str_format)
                    else:
                        ts_index = pd.to_datetime(dft.pop(ts_column))
                    dft.index = ts_index
                elif type(str_first_value) == pd.Timestamp or type(str_first_value) == np.datetime64:
                    ### if it is a datetime column, then set it as index
                    ### if it a datetime index, then just set the index as is 
                    ts_index = dft.pop(ts_column)
                    dft.index = ts_index
                elif type(str_first_value) in [np.int8, np.int16, np.int32, np.int64]:
                    ### if it is an integer column, convert ts_column into datetime and then set as index
                    ts_index = pd.to_datetime(dft.pop(ts_column))
                    dft.index = ts_index
                else:
                    print('    Type of time series column %s is float or unknown. Must be string or datetime. Please check input and try again.' %ts_column)
                    return 
            elif ts_column in dft.index.name:
                print('    train time series %s column is the index on test data...' %ts_column)
                ts_index = dft.index
                str_first_value = ts_index[0]
                str_values = ts_index[:12]
                if type(str_first_value) == str:
                    ### if index is in string format, you must infer its datetime string format and then set datetime index
                    str_format = infer_date_time_format(str_values)
                    if str_format:
                        str_format = str_format[0]
                        ts_index = pd.to_datetime(ts_index, format=str_format)
                    else:
                        ts_index = pd.to_datetime(ts_index)
                    dft.index = ts_index
                elif type(ts_index) == pd.core.indexes.datetimes.DatetimeIndex:
                    ### if dft already has a datetime index, leave it as it is
                    pass
                elif type(ts_index) == pd.DatetimeIndex or dft.index.dtype == '<M8[ns]':
                    ### if dft already has a datatime index, leave it as is
                    pass
                elif type(str_first_value) in [np.int8, np.int16, np.int32, np.int64]:
                    ### if it is not a datetime index, then convert it to datetime and set the index
                    ts_index = pd.to_datetime(ts_index)
                    dft.index = ts_index
                else:
                    print('    Type of index is unknown or float. It must be datetime or string. Please check input and try again.')
                    return
            else:
                print(f"    (Error) Cannot find '{ts_column}' (or index) in given data.")
                return None
        except:
            print('    Trying to convert time series column %s into index erroring. Please check input and try again.' %ts_column)
            return 
    elif type(dft) == dask.dataframe.core.DataFrame:
        str_format = ''
        if ts_column in dft.columns:
            print('    %s column exists in dask data frame...' %ts_column)
            str_first_value = dft[ts_column].compute()[0]
            dft.index = dd.to_datetime(dft[ts_column].compute())
            dft = dft.drop(ts_column, axis=1)
        elif ts_column in dft.index.name:
            print('    train index %s is already a time series index. Continuing...' %ts_column)
        else:
            print(f"    (Error) Model to be used for prediction 'ML'. Hence, input df must have a column (or index) called '{ts_column}' corresponding to the original ts_index column passed during training. No predictions will be made.")
            return None
    else:
        print('    Unable to detect type of data. Please check your input and try again')                        
        return
    return dft, str_format
############################################################################################################
def change_to_datetime_index_test(testdata, ts_column):
    testdata = copy.deepcopy(testdata)
    str_format = ''
    ##### This is where we change the time index of test data #############
    
    try:
        if isinstance(testdata, pd.Series) or isinstance(testdata, pd.DataFrame):
            if ts_column in testdata.columns:
                str_first_value = testdata[ts_column].values[0]
                str_values = testdata[ts_column].values[:12]
                if type(str_first_value) == str:
                    ### if ts_column is an object column, save its string format in date-time format
                    str_format = infer_date_time_format(str_values)
                    if str_format:
                        str_format = str_format[0]
                    else:
                        str_format = ''
                else:
                    ### If ts_column is not a string column, then set its format to an empty string ##
                    str_format = ''
                ###### If the str_format is detected, set the index as time series index ##
                ts_index = testdata.pop(ts_column)
                if str_format:
                    ts_index = pd.to_datetime(ts_index, format=str_format)
                else:
                    ts_index = pd.to_datetime(ts_index)
                testdata.index = ts_index
            elif ts_column in testdata.index.name:
                ts_index = testdata.index
                str_first_value = ts_index[0]
                str_values = ts_index[:12]
                if type(str_first_value) == str:
                    ### if index is in string format, you must infer its datetime string format and then set datetime index
                    str_format = infer_date_time_format(str_values)
                    if str_format:
                        str_format = str_format[0]
                    else:
                        str_format = ''
                else:
                    ### if index is in string format, you must infer its datetime string format and then set datetime index
                    str_format = ''
                ### now set the index to datetime format
                if str_format:
                    ts_index = pd.to_datetime(ts_index, format=str_format)
                else:
                    ts_index = pd.to_datetime(ts_index)
                dft.index = ts_index
        elif type(testdata) == dask.dataframe.core.DataFrame:
            #### the below tests work for a dask dataframe as well ##
            if ts_column in testdata.columns:
                str_first_value = testdata[ts_column].compute().values[0]
                str_values = testdata[ts_column].compute().values[:12]
                if type(str_first_value) == str:
                    ### if ts_column is an object column, save its string format in date-time format
                    str_format = infer_date_time_format(str_values)
                    if str_format:
                        str_format = str_format[0]
                    else:
                        str_format = ''
                else:
                    ### If ts_column is not a string column, then set its format to an empty string ##
                    str_format = ''
                ####### Now set the index to datetime index and drop the ts_colum #########
                testdata.index = dd.to_datetime(testdata[ts_column].compute())
                testdata = testdata.drop(ts_column, axis=1)
            elif ts_column in testdata.index.name:
                #### the above test works for a dask dataframe as well ##
                ts_index = testdata.index
                if type(testdata.index.compute().values[0]) in [np.datetime64]:
                    str_values = testdata.index.compute().values
                    str_values_dt = pd.to_datetime(str_values)
                    str_values = str_values_dt.astype(str)
                    str_first_value = str_values[0]
                    ## there is no need to change its index if the index is already datetime index ##
                    str_format = infer_date_time_format(str_values[:12])
                else:
                    str_first_value = ts_index[0]
                    str_values = ts_index.astype(str)
                    testdata.index = pd.to_datetime(str_values, format=str_format)
                    ### if index is in string format, you must infer its datetime string format and then set datetime index
                    str_format = infer_date_time_format(str_values[:12])
                if str_format:
                    str_format = str_format[0]
                else:
                    str_format = ''
            else:
                print("Error: Cannot detect %s either in columns or index. Please check input and try again." %ts_column)
        else:
            print('Unknown type of testdata. Please check input and try again.')
    except:
        print('    converting testdata to datetime index erroring. Please check input and try again.')
    #### this is where we return the testdata and format 
    return testdata, str_format

def time_series_split(ts_df):
    """
    This utility splits any dataframe sent as a time series split using the sklearn function.
    """
    tscv = TimeSeriesSplit(n_splits=2)
    train_index, test_index = list(tscv.split(ts_df))[1][0], list(tscv.split(ts_df))[1][1]
    ts_train, ts_test = ts_df[ts_df.index.isin(train_index)], ts_df[
                        ts_df.index.isin(test_index)]
    print(ts_train.shape, ts_test.shape)
    return ts_train, ts_test


def convert_timeseries_dataframe_to_supervised(df: pd.DataFrame, namevars, target, n_in=1, n_out=0, dropT=True):
    """
    Transform a time series in dataframe format into a supervised learning dataset while
    keeping dataframe intact.
    Returns the transformed pandas DataFrame, the name of the target column and the names of the predictor columns
    Arguments:
        df: A timeseries dataframe that you want to convert to Supervised dataset.
        namevars: columns that you want to lag in the data frame. Other columns will be untouched.
        target: this is the target variable you intend to use in supervised learning
        n_in: Number of lag periods as input (X).
        n_out: Number of future periods (optional) as output for the taget variable (y).
        dropT: Boolean - whether or not to drop columns at time 't'.
        Returns:
        df: This is the transformed data frame with the time series columns laggged.
        Note that the original columns are dropped if you set the 'dropT' argument to True.
        If not, they are preserved.
    This Pandas DataFrame of lagged time series data is immediately available for supervised learning.

    rtype: pd.DataFrame, str, List[str]
    """
    
    df = copy.deepcopy(df)
    int_vars  = df.select_dtypes(include='integer').columns.tolist()
    # Notice that we will create a sequence of columns from name vars with suffix (t-n,... t-1), etc.
    drops = []
    int_changes = []
    for i in range(n_in, -1, -1):
        if i == 0:
            for var in namevars:
                addname = var + '(t)'
                df = df.rename(columns={var:addname})
                drops.append(addname)
                if var in int_vars:
                    int_changes.append(addname)
        else:
            for var in namevars:
                addname = var + '(t-' + str(i) + ')'
                df[addname] = df[var].shift(i)
                if var in int_vars:
                    int_changes.append(addname)
    ## forecast sequence (t, t+1,... t+n)
    if n_out == 0:
        n_out = False
    for i in range(1, n_out):
        for var in namevars:
            addname = var + '(t+' + str(i) + ')'
            df[addname] = df[var].shift(-i)
    #	drop rows with NaN values
    df = df.dropna()

    ### Make sure that whatever vars came in as integers return back as integers!
    df[int_changes] = df[int_changes].astype(np.int64)

    #	put it all together
    df = df.rename(columns={target+'(t)':target})
    if dropT:
        ### If dropT is true, all the "t" series of the target column (in case it is in the namevars)
        ### will be removed if you don't want the target to learn from its "t" values.
        ### Similarly, we will also drop all the "t" series of name_vars if you set dropT to Trueself.
        try:
            drops.remove(target)
        except:
            pass
        df.drop(drops, axis=1, inplace=True)
    preds = [x for x in list(df) if x not in [target]]
    
    return df, target, preds
##############################################################################################
def find_max_min_value_in_a_dataframe(df, max_min='min'):
    """
    This returns the lowest or highest value in a df and its row value where it can be found.
    Unfortunately, it does not return the column where it is found. So not used much.
    """
    if max_min == 'min':
        return df.loc[:, list(df)].min(axis=1).min(), df.loc[:, list(df)].min(axis=1).idxmin()
    else:
        return df.loc[:, list(df)].max(axis=1).max(), df.loc[:, list(df)].min(axis=1).idxmax()
##############################################################################################
# THIS IS A MORE COMPLEX ALGORITHM THAT CHECKS MORE SPECIFICALLY FOR A DATE AND TIME FIELD
import datetime as dt
from datetime import datetime, date, time

### This tests if a string is date field and returns a date type object if successful and
##### a null list if it is unsuccessful
def is_date(txt):
    fmts = ('%Y-%m-%d', '%d/%m/%Y', '%d-%b-%Y', '%d/%b/%Y', '%b/%d/%Y', '%m/%d/%Y', '%b-%d-%Y', '%m-%d-%Y',
 '%Y/%m/%d', '%m/%d/%y', '%d/%m/%y', '%Y-%b-%d', '%Y-%B-%d', '%d-%m-%y', '%a, %d %b %Y', '%a, %d %b %y',
 '%d %b %Y', '%d %b %y', '%a, %d/%b/%y', '%d-%b-%y', '%m-%d-%y', '%d-%m-%Y', '%b%d%Y', '%d%b%Y',
 '%Y', '%b %d, %Y', '%B %d, %Y', '%B %d %Y', '%b %Y', '%B%Y', '%b %d,%Y')
    parsed=None
    for fmt in fmts:
        try:
            t = dt.datetime.strptime(txt, fmt)
            Year=t.year
            if Year > 2040 or Year < 1900:
                pass
            else:
                parsed = fmt
                return fmt
                break
        except ValueError as err:
            pass
    return parsed



#### This tests if a string is time field and returns a time type object if successful and
##### a null list if it is unsuccessful
def is_time(txt):
    fmts = ('%H:%M:%S.%f','%M:%S.%fZ','%Y-%m-%dT%H:%M:%S.%fZ','%h:%M:%S.%f','%-H:%M:%S.%f',
            '%H:%M','%I:%M','%H:%M:%S','%I:%M:%S','%H:%M:%S %p','%I:%M:%S %p',
           '%H:%M %p','%I:%M %p')
    parsed=None
    for fmt in fmts:
        try:
            t = dt.datetime.strptime(txt, fmt)
            parsed=fmt
            return parsed
            break
        except ValueError as err:
            pass
    return parsed

#### This tests if a string has both date and time in it. Returns a date-time object and null if it is not

def is_date_and_time(txt):
    fmts = ('%d/%m/%Y  %I:%M:%S %p', '%d/%m/%Y %I:%M:%S %p', '%d-%b-%Y %I:%M:%S %p',
 '%d/%b/%Y %I:%M:%S %p', '%b/%d/%Y %I:%M:%S %p', '%Y-%m-%dT%H:%M:%SZ', '%Y-%m-%dT%H:%M:%S.%fZ',
 '%m/%d/%Y %I:%M %p', '%m/%d/%Y %H:%M %p', '%d/%m/%Y  %I:%M:%S', '%d/%m/%Y  %H:%M', '%m/%d/%Y %H:%M',
 '%m/%d/%Y  %H:%M', '%d/%m/%Y  %I:%M', '%d/%m/%Y  %I:%M %p', '%m/%d/%Y  %I:%M', '%d/%b/%Y  %I:%M',
 '%b/%d/%Y  %I:%M', '%m/%d/%Y  %I:%M:%S', '%b-%d-%Y %I:%M:%S %p', '%m-%d-%Y %H:%M:%S %p',
 '%b-%d-%Y %H:%M:%S %p', '%m/%d/%Y %H:%M:%S %p', '%b/%d/%Y %H:%M:%S %p', '%Y-%m-%d %H:%M:%S %Z',
 '%Y-%m-%d %H:%M:%S %Z%z', '%Y-%m-%d %H:%M:%S %z', '%Y/%m/%d %H:%M:%S %Z%z', '%m/%d/%y %H:%M:%S %Z%z',
 '%d/%m/%Y %H:%M:%S %Z%z', '%m/%d/%Y %H:%M:%S %Z%z', '%d/%m/%y %H:%M:%S %Z%z', '%Y-%b-%d %H:%M:%S %Z%z',
 '%Y-%B-%d %H:%M:%S %Z%z', '%d-%b-%Y %H:%M:%S %Z%z', '%d-%m-%y %H:%M:%S %Z%z', '%Y-%m-%d %H:%M',
 '%Y-%b-%d %H:%M', '%a, %d %b %Y %T %z', '%a, %d %b %y %T %z', '%d %b %Y %T %z', '%d %b %y %T %z',
 '%d/%b/%Y %T %z', '%a, %d/%b/%y %T %z', '%d-%b-%Y %T %z', '%d-%b-%y %T %z', '%m-%d-%Y %I:%M %p',
 '%m-%d-%y %I:%M %p', '%m-%d-%Y %I:%M:%S %p', '%d-%m-%Y %H:%M:%S %p', '%m-%d-%y %H:%M:%S %p',
 '%d-%b-%Y %H:%M:%S %p', '%d-%m-%y %H:%M:%S %p', '%d-%b-%y %I:%M:%S %p', '%d-%b-%y %I:%M %p',
 '%d-%b-%Y %I:%M %p', '%d-%m-%Y %H:%M %p', '%d-%m-%y %H:%M %p')
    parsed=None
    for fmt in fmts:
        try:
            t = dt.datetime.strptime(txt, fmt)
            parsed=fmt
            return parsed
            break
        except ValueError as err:
            pass
    return parsed

# FIND DATE TIME VARIABLES

# This checks if a field in general is a date or time field

def infer_date_time_format(list_dates):
    """
    This is a generic algorithm that can infer date and time formats by checking repeatedly against a list.
    Make sure you give it a list of datetime formats since there can be many formats in a list.
    You can take the first of the returned list of formats or the majority or whatever you wish.
    # THE DATE FORMATS tested so far by this algorithm are:
        # 19JAN1990
        # JAN191990
        # 19/jan/1990
        # jan/19/1990
        # Jan 19, 1990
        # January 19, 1990
        # Jan 19,1990
        # 01/19/1990
        # 01/19/90
        # 1990
        # Jan 1990
        # January1990 
        # YOU CAN ADD MORE FORMATS above IN THE "fmts" section.
    """
    date_time_fmts = []
    try: 
        for each_datetime in list_dates:
            date1 = is_date(each_datetime)
            if date1 and not date1 in date_time_fmts:
                date_time_fmts.append(date1)
            else:
                date2 = is_time(each_datetime)
                if date2 and not date2 in date_time_fmts:
                    date_time_fmts.append(date2)
                else:
                    date3 = is_date_and_time(each_datetime)
                    if date3 and not date3 in date_time_fmts:
                        date_time_fmts.append(date3)
            if not date1 and not date2 and not date3 :
                print('date time format cannot be inferred. Please check input and try again.')
    except:
        print('Error in inferring date time format. Returning...')
    return date_time_fmts
#################################################################################################
def reduce_mem_usage(df):
    """
    #####################################################################
    Greatly indebted to :
    https://www.kaggle.com/arjanso/reducing-dataframe-memory-size-by-65
        for this function to reduce memory usage.
    #####################################################################
    It is a bit slow as it iterate through all the columns of a dataframe and modifies data types
        to reduce memory usage. But it has been shown to reduce memory usage by 65% or so.       
    """
    start_mem = df.memory_usage().sum() / 1024**2
    if type(df) == dask.dataframe.core.DataFrame:
        start_mem = start_mem.compute()
    print('    Caution: We will try to reduce the memory usage of dataframe from {:.2f} MB'.format(start_mem))
    cols = df.columns
    if type(df) == dask.dataframe.core.DataFrame:
        cols = cols.tolist()
    datevars = df.select_dtypes(include=[np.datetime64]).columns.tolist()
    numvars = df.select_dtypes(include='number').columns.tolist()
    for col in cols:
        col_type = df[col].dtype
        if col in datevars:
            pass
        elif col in numvars:
            c_min = df[col].min()
            c_max = df[col].max()
            if type(df) == dask.dataframe.core.DataFrame:
                c_min = c_min.compute()
                c_max = c_max.compute()
            if str(col_type)[:3] == 'int':
                if c_min > np.iinfo(np.int8).min and c_max < np.iinfo(np.int8).max:
                    df[col] = df[col].astype(np.int8)
                elif c_min > np.iinfo(np.int16).min and c_max < np.iinfo(np.int16).max:
                    df[col] = df[col].astype(np.int16)
                elif c_min > np.iinfo(np.int32).min and c_max < np.iinfo(np.int32).max:
                    df[col] = df[col].astype(np.int32)
                elif c_min > np.iinfo(np.int64).min and c_max < np.iinfo(np.int64).max:
                    df[col] = df[col].astype(np.int64)  
            else:
                if c_min > np.finfo(np.float16).min and c_max < np.finfo(np.float16).max:
                    df[col] = df[col].astype(np.float16)
                elif c_min > np.finfo(np.float32).min and c_max < np.finfo(np.float32).max:
                    df[col] = df[col].astype(np.float32)
                else:
                    df[col] = df[col].astype(np.float64)
        else:
            df[col] = df[col].astype('category')

    #######  Results after memory usage function ###################
    end_mem = df.memory_usage().sum() / 1024**2
    if type(df) == dask.dataframe.core.DataFrame:
        end_mem = end_mem.compute()
    print('    Memory usage after optimization is: {:.2f} MB'.format(end_mem))
    print('        decreased by {:.1f}%'.format(100 * (start_mem - end_mem) / start_mem))
    
    return df
##################################################################################


"""Module to Build a Prphet Model
"""
from typing import Optional
import logging
import copy
import time
import numpy as np
#from tscv import GapWalkForward # type: ignore
from sklearn.model_selection import TimeSeriesSplit

import pandas as pd # type: ignore
from pandas.core.generic import NDFrame # type:ignore
import pdb
import dask
import dask.dataframe as dd

import matplotlib.pyplot as plt # type: ignore

from fbprophet import Prophet # type: ignore
from fbprophet.diagnostics import cross_validation
from fbprophet.diagnostics import performance_metrics
from fbprophet.plot import plot_cross_validation_metric

from .build_base import BuildBase

# helper functions
from ..utils import print_dynamic_rmse, quick_ts_plot, print_ts_model_stats
from ..utils.logging import SuppressStdoutStderr

#### Suppress INFO messages from FB Prophet!

logging.getLogger('fbprophet').setLevel(logging.WARNING)


class BuildProphet(BuildBase):
    """Class to build a Prophet Model
    """
    def __init__(
            self, forecast_period, time_interval, seasonal_period, scoring, verbose,
             conf_int, holidays, growth, seasonality, **kwargs
        ):
        """
        Automatically build a Prophet Model
        """
        super().__init__(
            scoring=scoring,
            forecast_period=forecast_period,
            verbose=verbose
        )
        self.time_interval = time_interval
        self.seasonal_period = seasonal_period
        self.conf_int = conf_int
        self.holidays = holidays
        self.growth = growth
        self.seasonality = seasonality
        yearly_seasonality = False
        daily_seasonality = False
        weekly_seasonality = False
        if self.time_interval == 'weeks':
            weekly_seasonality =  seasonality
        elif self.time_interval == 'years':
            yearly_seasonality = seasonality
        elif self.time_interval == 'days':
            daily_seasonality = seasonality
        #self.model = Prophet(
        #    yearly_seasonality=yearly_seasonality,
        #    weekly_seasonality=weekly_seasonality,
        #    daily_seasonality=daily_seasonality,
        #    interval_width=self.conf_int,
        #    holidays = self.holidays,
        #    growth = self.growth)
        self.model = Prophet(growth = self.growth)
        self.univariate = None
        self.list_of_valid_time_ints = ['B','C','D','W','M','SM','BM','CBM',
                                        'MS','SMS','BMS','CBMS','Q','BQ','QS','BQS',
                                        'A,Y','BA,BY','AS,YS','BAS,BYS','BH',
                                        'H','T,min','S','L,ms','U,us','N']
        self.list_of_valid_time_ints.append(time_interval)

        if kwargs:
            for key, value in zip(kwargs.keys(),kwargs.values()):
                if key == 'seasonality_mode':
                    self.seasonality = True
                    key = value
                else:
                    key = value

    def fit(self, ts_df: pd.DataFrame, target_col: str, cv: Optional[int], time_col: str):
        """
        Fits the model to the data

        :param ts_df The time series data to be used for fitting the model
        :type ts_df pd.DataFrame

        :param target_col The column name of the target time series that needs to be modeled.
        All other columns will be considered as exogenous variables (if applicable to method)
        :type target_col str

        :param cv: Number of folds to use for cross validation.
        Number of observations in the Validation set for each fold = forecast period
        If None, a single fold is used
        :type cv Optional[int]

        :param time_col: Name of the time column in the dataset (needed by Prophet)
        Time column can also be the index, in which case, this would be the name of the index
        :type time_col str

        :rtype object
        """
        # use all available threads/cores

        self.time_col = time_col
        self.original_target_col = target_col
        self.original_preds = [x for x in list(ts_df) if x not in [self.original_target_col]]

        if len(self.original_preds) == 0:
            self.univariate = True
        else:
            self.univariate = False

        # print(f"Prophet Is Univariate: {self.univariate}")

        ts_df = copy.deepcopy(ts_df)

        ##### if you are going to use matplotlib with prophet data, it gives an error unless you do this.
        pd.plotting.register_matplotlib_converters()

        #### You have to import Prophet if you are going to build a Prophet model #############
        actual = 'y'
        timecol = 'ds'

        data = self.prep_col_names_for_prophet(ts_df=ts_df, test=False)

        if self.univariate:
            dft = data[[timecol, actual]]
        else:
            dft = data[[timecol, actual] + self.original_preds]

        ##### For most Financial time series data, 80 percent conf interval is enough...
        if self.verbose >= 1:
            print('    Fit-Predict data (shape=%s) with Confidence Interval = %0.2f...' %(dft.shape, self.conf_int))
        ### Make Sure you lower your desired interval width from the normal 95% to a more realistic 80%
        start_time = time.time()

        if self.univariate is False:
            for name in self.original_preds:
                self.model.add_regressor(name)

        print("  Starting Prophet Fit")

        if self.seasonality:
            prophet_seasonality, prophet_period, fourier_order, prior_scale = get_prophet_seasonality(
                                        self.time_interval, self.seasonal_period)
            self.model.add_seasonality(name=prophet_seasonality,
                            period=prophet_period, fourier_order=fourier_order, prior_scale= prior_scale)
            print('       Adding %s seasonality to Prophet with period=%d, fourier_order=%d and prior_scale=%0.2f' %(
                            prophet_seasonality, prophet_period, fourier_order, prior_scale))
        else:
            print('      No seasonality assumed since seasonality flag is set to False')


        if type(dft) == dask.dataframe.core.DataFrame:
            num_obs = dft.shape[0].compute()
        else:
            num_obs = dft.shape[0]

        ### Creating a new way to skip cross validation when trying to run auto-ts multiple times. ###
        if cv == 0:
            cv_in = 0
        else:
            cv_in = copy.deepcopy(cv)
        NFOLDS = self.get_num_folds_from_cv(cv)

        #########################################################################################
        # NOTE: This change to the FB recommendation will cause the cv folds from facebook to
        # be incompatible with the folds from the other models (in terms of periods of evaluation
        # as well as number of observations in each period). Hence the final comparison will
        # be biased since it will not compare the same folds.

        # The original implementation was giving issues under certain conditions, hence this change
        # to FB recommendation has been made as a temporary (short term) fix.
        # The root cause issue will need to be fixed eventually at a later point.
        #########################################################################################

        ### Prophet's Time Interval translates into frequency based on the following pandas date_range alias:
        #  Link: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timeseries-offset-aliases
        ## This is done using the get_prophet_time_interval() function later.
        
        if self.time_interval in self.list_of_valid_time_ints:
            time_int = copy.deepcopy(self.time_interval)
        else:
            time_int = self.get_prophet_time_interval(for_cv=False)
        
        # First  Fold -->
        #   Train Set: 0:initial
        #   Test Set: initial:(initial+horizon)
        # Second Fold -->
        #   Train Set: (period):(initial+period)
        #   Test Set: (initial+period):(initial+horizon+ period)
        # Format: '850 D'

        print("  Starting Prophet Cross Validation")
        ################################################################################
        if self.forecast_period <= 5:
            #### Set a minimum of 5 for the number of rows in test!
            self.forecast_period = 5
        ### In case the number of forecast_period is too high, just reduce it so it can fit into num_obs
        if NFOLDS*self.forecast_period > num_obs:
            self.forecast_period = int(num_obs/(NFOLDS+1))
            print('Lowering forecast period to %d to enable cross_validation' %self.forecast_period)
        ###########################################################################################
        #cv = GapWalkForward(n_splits=NFOLDS, gap_size=0, test_size=self.forecast_period)
        max_trainsize = len(dft) - self.forecast_period
        try:
            cv = TimeSeriesSplit(n_splits=NFOLDS, test_size=self.forecast_period) ### this works only sklearn v 0.0.24]
        except:
            cv = TimeSeriesSplit(n_splits=NFOLDS, max_train_size = max_trainsize)
        y_preds = pd.DataFrame()
        print('Max. iterations using expanding window cross validation = %d' %NFOLDS)
        start_time = time.time()
        rmse_folds = []
        norm_rmse_folds = []
        forecast_df_folds = []

        concatenated = pd.DataFrame()
        extra_concatenated = pd.DataFrame()

        if type(dft) == dask.dataframe.core.DataFrame:
            dft = dft.head(len(dft)) ### this converts dask into a pandas dataframe

        if  cv_in == 0:
            print('Skipping cross validation steps since cross_validation = %s' %cv_in)
            model = Prophet(growth="linear")
            kwargs = {'iter':1e2} ## this limits iterations and hence speeds up prophet
        else:
            for fold_number, (train_index, test_index) in enumerate(cv.split(dft)):
                dftx = dft.head(len(train_index)+len(test_index))
                train_fold = dftx.head(len(train_index)) ## now train will be the first segment of dftx
                test_fold = dftx.tail(len(test_index)) ### now test will be right after train in dftx

                horizon = len(test_fold)
                print(f"\nFold Number: {fold_number+1} --> Train Shape: {train_fold.shape[0]} Test Shape: {test_fold.shape[0]}")

                #########################################
                #### Define the model with fold data ####
                #########################################

                model = Prophet(growth="linear")

                ############################################
                #### Fit the model with train_fold data ####
                ############################################

                kwargs = {'iter':1e2} ## this limits iterations and hence speeds up prophet
                model.fit(train_fold, **kwargs)

                #################################################
                #### Predict using model with test_fold data ####
                #################################################
                
                future_period = model.make_future_dataframe(freq=time_int, periods=horizon)
                forecast_df = model.predict(future_period)
                ### Now compare the actuals with predictions ######

                y_pred = forecast_df['yhat'][-horizon:]

                concatenated = pd.DataFrame(np.c_[test_fold[actual].values,
                            y_pred.values], columns=['original', 'predicted'],index=test_fold.index)

                if fold_number == 0:
                    extra_concatenated = copy.deepcopy(concatenated)
                else:
                    extra_concatenated = extra_concatenated.append(concatenated)

                rmse_fold, rmse_norm = print_dynamic_rmse(concatenated['original'].values, concatenated['predicted'].values,
                                            concatenated['original'].values)

                print('Cross Validation window: %d completed' %(fold_number+1,))
                rmse_folds.append(rmse_fold)
                norm_rmse_folds.append(rmse_norm)

            ######################################################
            ### This is where you consolidate the CV results #####
            ######################################################
            fig = model.plot(forecast_df)
            #rmse_mean = np.mean(rmse_folds)
            #print('Average CV RMSE over %d windows (macro) = %0.5f' %(fold_number+1,rmse_mean))


            #cv_micro = np.sqrt(mean_squared_error(y_trues.values, y_preds.values))
            #print('Average CV RMSE of all predictions (micro) = %0.5f' %cv_micro)

            try:
                print_ts_model_stats(extra_concatenated['original'], extra_concatenated['predicted'], "Prophet")
            except:
                print('Error: Not able to plot Prophet CV results')

            forecast_df_folds = extra_concatenated['predicted'].values
            #print("  End of Prophet Cross Validation")
            print('Time Taken = %0.0f seconds' %((time.time()-start_time)))

        #### Now you need to fit Prophet on the whole train data set ##########
        dftx = dft.head(len(dft))
        model = Prophet(growth="linear")
        self.model = model
        self.model.fit(dftx, **kwargs)
        print("  End of Prophet Fit")

        #num_obs_folds = df_cv.groupby('cutoff')['ds'].count()

        # https://stackoverflow.com/questions/54405704/check-if-all-values-in-dataframe-column-are-the-same
        #a = num_obs_folds.to_numpy()
        #all_equal = (a[0] == a).all()

        #if not all_equal:
            #print("WARNING: All folds did not have the same number of observations in the validation sets.")
            #print("Num Test Obs Per fold")
            #print(num_obs_folds)

        #rmse_folds = []
        #norm_rmse_folds = []
        #forecast_df_folds = []

        #df_cv_grouped = df_cv.groupby('cutoff')
        #for (_, loop_df) in df_cv_grouped:
        #    rmse, norm_rmse = print_dynamic_rmse(loop_df['y'], loop_df['yhat'], dft['y'])
        #    rmse_folds.append(rmse)
        #    norm_rmse_folds.append(norm_rmse)
        #    forecast_df_folds.append(loop_df)

        # print(f"RMSE Folds: {rmse_folds}")
        # print(f"Norm RMSE Folds: {norm_rmse_folds}")
        # print(f"Forecast DF folds: {forecast_df_folds}")

        # forecast = self.predict(simple=False, return_train_preds=True)

        # ####  We are going to plot Prophet's forecasts differently since it is better
        # dfa = plot_prophet(dft, forecast);
        # # Prophet makes Incredible Predictions Charts!
        # ###  There can't be anything simpler than this to make Forecasts!
        # #self.model.plot(forecast);  # make sure to add semi-colon in the end to avoid plotting twice
        # # Also their Trend, Seasonality Charts are Spot On!
        # try:
        #     self.model.plot_components(forecast)
        # except:
        #     print('Error in FB Prophet components forecast. Continuing...')

        #rmse, norm_rmse = print_dynamic_rmse(dfa['y'], dfa['yhat'], dfa['y'])

        #return self.model, forecast, rmse, norm_rmse
        return self.model, forecast_df_folds, rmse_folds, norm_rmse_folds

    def refit(self, ts_df: pd.DataFrame) -> object:
        """
        Refits an already trained model using a new dataset
        Useful when fitting to the full data after testing with cross validation
        :param ts_df The time series data to be used for fitting the model
        :type ts_df pd.DataFrame
        :rtype object
        """

    def predict(
            self,
            testdata: Optional[pd.DataFrame] = None,
            forecast_period: Optional[int] = None,
            simple: bool = False,
            return_train_preds: bool = False) -> Optional[NDFrame]:
        """
        Return the predictions
        :param testdata The test dataframe containing the exogenous variables to be used for prediction.
        :type testdata Optional[pd.DataFrame]
        :param forecast_period The number of periods to make a prediction for.
        :type forecast_period Optional[int]
        :param simple If True, this method just returns the predictions.
        If False, it will return the standard error, lower and upper confidence interval (if available)
        :type simple bool
        :param return_train_preds If True, this method just returns the train predictions along with test predictions.
        If False, it will return only test predictions
        :type return_train_preds bool
        :rtype NDFrame
        """

        """
        Return the predictions
        # TODO: What about future exogenous variables?
        # https://towardsdatascience.com/forecast-model-tuning-with-additional-regressors-in-prophet-ffcbf1777dda
        """

        # if testdata is not None:
        #     warnings.warn(
        #         "Multivariate models are not supported by the AutoML prophet module." +
        #         "Univariate predictions will be returned for now."
        #     )

        # Prophet is a Little Complicated - You need 2 steps to Forecast
        ## 1. You need to create a dataframe to hold the predictions which specifies datetime
        ##    periods that you want to predict. It automatically creates one with both past
        ##    and future dates.
        ## 2. You need to ask Prophet to make predictions for the past and future dates in
        ##    that dataframe above.
        ## So if you had 2905 rows of data, and ask Prophet to predict for 365 periods,
        ##    it will give you predictions of the past (2905) and an additional 365 rows
        ##    of future (total: 3270) rows of data.
        ### This is where we take the first steps to make a forecast using Prophet:
        ##   1. Create a dataframe with datetime index of past and future dates
        
        # Next we ask Prophet to make predictions for those dates in the dataframe along with prediction intervals
        if self.time_interval in self.list_of_valid_time_ints:
            time_int = copy.deepcopy(self.time_interval)
        else:
            time_int = self.get_prophet_time_interval(for_cv=False)

        if self.univariate:
            if isinstance(testdata, int):
                forecast_period = testdata
            elif isinstance(testdata, pd.DataFrame):
                forecast_period = testdata.shape[0]
                if testdata.shape[0] != self.forecast_period:
                    self.forecast_period = testdata.shape[0]
            else:
                forecast_period = self.forecast_period
            self.forecast_period = forecast_period
            future = self.model.make_future_dataframe(periods=self.forecast_period, freq=time_int)
        else:
            if isinstance(testdata, int) or testdata is None:
                print("(Error): Model is Multivariate, hence test dataframe must be provided for prediction.")
                return None
            elif isinstance(testdata, pd.DataFrame):
                forecast_period = testdata.shape[0]
                if testdata.shape[0] != self.forecast_period:
                    self.forecast_period = testdata.shape[0]
                future = self.prep_col_names_for_prophet(ts_df=testdata, test=True)
        print('Building Forecast dataframe. Forecast Period = %d' % self.forecast_period)
        ### This will work in both univariate and multi-variate cases now ######

        forecast = self.model.predict(future)

        # Return values for the forecast period only
        if simple:
            if return_train_preds:
                forecast = forecast['yhat']
            else:
                if forecast_period is None:
                    forecast = forecast['yhat']
                else:
                    forecast = forecast.iloc[-forecast_period:]['yhat']

        else:
            if return_train_preds:
                forecast = forecast
            else:
                if forecast_period is None:
                    forecast = forecast['yhat']
                else:
                    forecast = forecast.iloc[-forecast_period:]

        return forecast

    # TODO: Update: This method will not be used in CV since it is in D always.
    # Hence Remove the 'for_cv' argument
    def get_prophet_time_interval(self, for_cv: bool = False) -> str:
        """
        Returns the time interval in Prophet compatible format

        :param for_cv If False, this will return the format needed to make future dataframe (for univariate analysis)
        If True, this will return the format needed to be passed to the cross-validation object
        """
        if self.time_interval in ['months', 'month', 'm']:
            time_int = 'M'
        elif self.time_interval in ['days', 'daily', 'd']:
            time_int = 'D'
        elif self.time_interval in ['weeks', 'weekly', 'w']:
            time_int = 'W'
        # TODO: Add time_int for other options if they are different for CV and for future forecasts
        elif self.time_interval in ['qtr', 'quarter', 'q']:
            time_int = 'Q'
        elif self.time_interval in ['years', 'year', 'annual', 'y', 'a']:
            time_int = 'Y'
        elif self.time_interval in ['hours', 'hourly', 'h']:
            time_int = 'H'
        elif self.time_interval in ['minutes', 'minute', 'min', 'n']:
            time_int = 'M'
        elif self.time_interval in ['seconds', 'second', 'sec', 's']:
            time_int = 'S'
        else:
            time_int = 'W'
        return time_int

    def prep_col_names_for_prophet(self, ts_df: pd.DataFrame, test: bool = False) -> pd.DataFrame:
        """
        Renames the columns of the input dataframe to the right format needed by Prophet
        Target is renamed to 'y' and the time column is renamed to 'ds'
        # TODO: Complete docstring
        """

        if self.time_col not in ts_df.columns:
            #### This happens when time_col is not found but it's actually the index. In that case, reset index
            data = ts_df.reset_index()
        else:
            data = ts_df.copy(deep=True)

        if self.time_col not in data.columns:
            print("(Error): You have not provided the time_column values. This will result in an error")

        if test is False:
            data = data.rename(columns={self.time_col: 'ds', self.original_target_col: 'y'})
        else:
            data = data.rename(columns={self.time_col: 'ds'})

        return data

def plot_prophet(dft, forecastdf):
    """
    This is a different way of plotting Prophet charts as described in the following article:
    Source: https://nextjournal.com/viebel/forecasting-time-series-data-with-prophet
    Reproduced with gratitude to the author.
    """
    dft = copy.deepcopy(dft)
    forecastdf = copy.deepcopy(forecastdf)
    dft.set_index('ds', inplace=True)
    forecastdf.set_index('ds', inplace=True)
    dft.index = pd.to_datetime(dft.index)
    connect_date = dft.index[-2]
    mask = (forecastdf.index > connect_date)
    predict_df = forecastdf.loc[mask]
    viz_df = dft.join(predict_df[['yhat', 'yhat_lower', 'yhat_upper']], how='outer')
    _, ax1 = plt.subplots(figsize=(20, 10))
    ax1.plot(viz_df['y'], color='red')
    ax1.plot(viz_df['yhat'], color='green')
    ax1.fill_between(viz_df.index, viz_df['yhat_lower'], viz_df['yhat_upper'],
                     alpha=0.2, color="darkgreen")
    ax1.set_title('Actual (Red) vs Forecast (Green)')
    ax1.set_ylabel('Values')
    ax1.set_xlabel('Date Time')
    plt.show(block=False)
    return viz_df
#################################
from sklearn.metrics import mean_squared_error
from fbprophet import Prophet
import time
import pdb
import copy
import matplotlib.pyplot as plt
def easy_cross_validation(train, target, initial, horizon, period):
    n_folds = int(((train.shape[0]-initial)/period)-1)
    y_preds = pd.DataFrame()
    print('Max. iterations using sliding window cross validation = %d' %n_folds)
    start_time = time.time()
    start_p = 0  ## this represents start of train fold
    end_p = initial  ## this represents end of train fold
    start_s = initial  ## this represents start of test fold
    end_s = initial + horizon ### this represents end of test fold
    rmse_means = []
    norm_rmse_means = []
    y_trues = pd.DataFrame()
    for i in range(n_folds):
        #start_p += i*period
        end_p += i*period
        train_fold = train[start_p:end_p]
        start_s += i*period
        end_s += i*period
        test_fold = train[start_s: end_s]
        if len(test_fold) == 0:
            break
        model = Prophet(growth="linear")
        kwargs = {'iter':1e2} ## this limits iterations and hence speeds up prophet
        model.fit(train_fold, **kwargs)
        future_period = model.make_future_dataframe(freq="MS",periods=horizon)
        forecast_df = model.predict(future_period)
        y_pred = forecast_df.iloc[start_s:end_s]['yhat']
        if i == 0:
            y_preds = copy.deepcopy(y_pred)
        else:
            y_preds = y_preds.append(y_pred)
        rmse_fold, rmse_norm = print_dynamic_rmse(test_fold[target],y_pred,test_fold[target])
        print('Cross Validation window: %d completed' %(i+1,))
        rmse_means.append(rmse_fold)
        norm_rmse_means.append(rmse_norm)
    ### This is where you consolidate the CV results ####
    #print('Time Taken = %0.0f mins' %((time.time()-start_time)/60))
    rmse_mean = np.mean(rmse_means)
    #print('Average CV RMSE over %d windows (macro) = %0.5f' %(i,rmse_mean))
    y_trues = train[-y_preds.shape[0]:][target]
    cv_micro = np.sqrt(mean_squared_error(y_trues.values,
                                          y_preds.values))
    #print('Average CV RMSE of all predictions (micro) = %0.5f' %cv_micro)
    try:
        quick_ts_plot(train[target], y_preds[-horizon:])
    except:
        print('Error: Not able to plot Prophet CV results')
    return y_trues, y_preds, rmse_means, norm_rmse_means
##################################################################################
def get_prophet_seasonality(time_int, seasonal_period):
    """
    This returns the prophet seasonality if sent in the time interval.
    """
    prophet_seasonality = None
    if seasonal_period is not None:
        prophet_period = seasonal_period
    if time_int in [ 'MS', 'M', 'SM', 'BM', 'CBM', 'SMS', 'BMS']:
        prophet_seasonality = 'monthly'
        if seasonal_period is None:
            prophet_period = 30.5
        fourier_order = 12
        prior_scale = 0.1
    elif time_int in ['D', 'B', 'C']:
        prophet_seasonality = 'daily'
        if seasonal_period is None:
            prophet_period = 1
        fourier_order = 15
        prior_scale = 0.1
    elif time_int in ['W']:
        prophet_seasonality = 'weekly'
        if seasonal_period is None:
            prophet_period = 7
        fourier_order = 20
        prior_scale = 0.1
    elif time_int in ['Q', 'BQ', 'QS', 'BQS']:
        prophet_seasonality = 'quarterly'
        if seasonal_period is None:
            prophet_period = 365.25/4
        fourier_order = 5
        prior_scale = 0.1
    elif time_int in ['A,Y', 'BA,BY', 'AS,YS', 'BAS,YAS']:
        prophet_seasonality = 'yearly'
        if seasonal_period is None:
            prophet_period = 365.25
        fourier_order = 5
        prior_scale = 0.1
    elif time_int in ['BH', 'H', 'h']:
        prophet_seasonality = 'hourly'
        if seasonal_period is None:
            prophet_period = 24
        fourier_order = 5
        prior_scale = 0.1
    elif time_int in ['T,min']:
        prophet_seasonality = 'hourly'
        if seasonal_period is None:
            prophet_period = 24
        fourier_order = 12
        prior_scale = 0.1
    elif time_int in ['S', 'L,milliseconds', 'U,microseconds', 'N,nanoseconds']:
        prophet_seasonality = 'hourly'
        if seasonal_period is None:
            prophet_period = 24
        fourier_order = 12
        prior_scale = 0.1
    else:
        #### Monthly is the default ###
        prophet_seasonality = 'monthly'
        if seasonal_period is None:
            prophet_period = 30.5
        fourier_order = 12
        prior_scale = 0.1
    return prophet_seasonality, prophet_period, fourier_order, prior_scale
#################################################################################



"""
Unit Tests for VAR Models

----------------------
Total Combinations: 4
----------------------
Seasonality: NA
Univariate, Multivariate: Simple Independent Test for Univariate (1)
CV: Yes, No (2)
"""

import sys
import os
import unittest
import math
import numpy as np # type: ignore
import pandas as pd # type: ignore

from pandas.testing import assert_series_equal # type: ignore
from pandas.testing import assert_frame_equal # type: ignore
from statsmodels.tsa.statespace.sarimax import SARIMAXResultsWrapper  # type: ignore

sys.path.append(os.environ['DEV_AUTOTS'])
from auto_ts import auto_timeseries as ATS

class TestVAR(unittest.TestCase):

    def setUp(self):
        # Pre Release


        datapath = 'example_datasets/'
        filename1 = 'Sales_and_Marketing.csv'
        dft = pd.read_csv(datapath + filename1, index_col = None)

        self.ts_column = 'Time Period'
        self.sep = ','
        self.target = 'Sales'
        self.preds = [x for x in list(dft) if x not in [self.ts_column, self.target]] # Exogenous variable names

        self.train_multivar = dft[:40]
        self.test_multivar = dft[40:]

        self.train_univar = dft[:40][[self.ts_column, self.target]]
        self.test_univar = dft[40:][[self.ts_column, self.target]]

        self.forecast_period = 8

        self.expected_pred_col_names = np.array(['mean', 'mean_se', 'mean_ci_lower', 'mean_ci_upper'])

        ########################
        #### Golden Results ####
        ########################

        # TODO: Add to each individual test
        ## For each of the 8 combinations, we need the following
        # Internal Validation results (for each fold)
        # Internal Validation RMSE (overall and for each fold)

        # External Test results (various combinations of prediction windows - same as forecast period OR not same)
        # External Test RMSE


        ############################
        #### VAR Golden Results ####
        ############################

        #### UNIVARIATE ####
        self.forecast_gold_var_univar = None
        self.rmse_gold_var_univar = math.inf
        self.forecast_gold_var_univar_series = None
        self.forecast_gold_var_univar_series_10 = None

        #### MULTIVARIATE ####

        # Internal (to AutoML) validation set results
        self.forecast_gold_var_multivar_internal_val_cv_fold1 = np.array([
            510.302336, 531.109224, 536.878513, 534.311164,
            529.305887, 525.199071, 523.015255, 522.445215
        ])

        self.forecast_gold_var_multivar_internal_val_cv_fold2 = np.array([
            741.377909, 676.233419, 615.538721, 571.797729,
            546.952783, 537.342231, 537.474487, 542.307393
        ])

        self.rmse_gold_var_multivar_cv_fold1 = 155.21757611
        self.rmse_gold_var_multivar_cv_fold2 = 112.4770318 # Without CV gets this result

        ## External Test Set results
        results = [
            675.899931, 622.204059, 578.38291, 553.067517,
            543.612945, 543.696406, 547.604403, 551.762352
            ]
        index = pd.to_datetime([
            '2014-05-01', '2014-06-01', '2014-07-01', '2014-08-01',
            '2014-09-01', '2014-10-01', '2014-11-01', '2014-12-01'
            ])
        self.forecast_gold_var_multivar = np.array(results)

        self.forecast_gold_var_multivar_series = pd.Series(data=results, index=index)
        self.forecast_gold_var_multivar_series.name = 'mean'

        results = results + [554.643756, 556.055009]
        index = pd.to_datetime([
            '2014-05-01', '2014-06-01', '2014-07-01', '2014-08-01',
            '2014-09-01', '2014-10-01', '2014-11-01', '2014-12-01',
            '2015-01-01', '2015-02-01'
            ])
        self.forecast_gold_var_multivar_series_10 = pd.Series(data=results, index=index)
        self.forecast_gold_var_multivar_series_10.name = 'mean'


    def test_noCV(self):
        """
        Test 1: VAR without CV
        """
        print("\n\n" + "*"*50)
        print("Performing Unit Test: 'test_noCV'")
        print("*"*50 + "\n\n")

        automl_model = ATS(
            score_type='rmse', forecast_period=self.forecast_period, time_interval='Month',
            non_seasonal_pdq=None, seasonality=False, seasonal_period=12,
            model_type='VAR',
            verbose=0)
        automl_model.fit(
            traindata=self.train_multivar,
            ts_column=self.ts_column,
            target=self.target,
            cv=None,
            sep=self.sep)

        ml_dict = automl_model.get_ml_dict()

        ######################
        ## External Results ##
        ######################

        # Simple forecast with forecast window = the one used in training
        # Using named model
        test_predictions = automl_model.predict(
            testdata=self.test_multivar[[self.ts_column] + self.preds], # Not needed for VAR
            forecast_period=self.forecast_period,
            model="VAR"
        )
        assert_series_equal(
            test_predictions['mean'].round(6),
            self.forecast_gold_var_multivar_series
        )

        # Simple forecast with forecast window != the one used in training
        # Using named model
        test_predictions = automl_model.predict(
            testdata=self.test_multivar[[self.ts_column] + self.preds], # Not needed for VAR
            forecast_period=10,
            model="VAR"
        )
        assert_series_equal(test_predictions['mean'].round(6), self.forecast_gold_var_multivar_series_10)

        # Complex forecasts (returns confidence intervals, etc.)
        test_predictions = automl_model.predict(
            testdata=self.test_multivar[[self.ts_column] + self.preds], # Not needed for VAR
            forecast_period=self.forecast_period,
            model="VAR",
            simple=False
        )
        self.assertIsNone(
            np.testing.assert_array_equal(
                test_predictions.columns.values, self.expected_pred_col_names
            )
        )

        ###################
        ## ML Dictionary ##
        ###################
        self.assertIsNone(
            np.testing.assert_array_equal(
                np.round(ml_dict.get('VAR').get('forecast')[0]['mean'].values.astype(np.double), 6),
                self.forecast_gold_var_multivar_internal_val_cv_fold2
            ),
            "(Multivar Test) VAR Forecast does not match up with expected values."
        )

        self.assertEqual(
            round(ml_dict.get('VAR').get('rmse')[0], 8), self.rmse_gold_var_multivar_cv_fold2,
            "(Multivar Test) VAR RMSE does not match up with expected values.")

    def test_CV(self):
        """
        Test 2: VAR with CV
        """
        print("\n\n" + "*"*50)
        print("Performing Unit Test: 'test_CV'")
        print("*"*50 + "\n\n")

        automl_model = ATS(
            score_type='rmse', forecast_period=self.forecast_period, time_interval='Month',
            non_seasonal_pdq=None, seasonality=False, seasonal_period=12,
            model_type='VAR',
            verbose=0)
        automl_model.fit(
            traindata=self.train_multivar,
            ts_column=self.ts_column,
            target=self.target,
            cv=2,
            sep=self.sep)

        ml_dict = automl_model.get_ml_dict()

        ######################
        ## External Results ##
        ######################

        # Simple forecast with forecast window = the one used in training
        # Using named model
        test_predictions = automl_model.predict(
            testdata=self.test_multivar[[self.ts_column] + self.preds], # Not needed for VAR
            forecast_period=self.forecast_period,
            model="VAR"
        )
        assert_series_equal(
            test_predictions['mean'].round(6),
            self.forecast_gold_var_multivar_series
        )

        # Simple forecast with forecast window != the one used in training
        # Using named model
        test_predictions = automl_model.predict(
            testdata=self.test_multivar[[self.ts_column] + self.preds], # Not needed for VAR
            forecast_period=10,
            model="VAR"
        )
        assert_series_equal(test_predictions['mean'].round(6), self.forecast_gold_var_multivar_series_10)

        # Complex forecasts (returns confidence intervals, etc.)
        test_predictions = automl_model.predict(
            testdata=self.test_multivar[[self.ts_column] + self.preds], # Not needed for VAR
            forecast_period=self.forecast_period,
            model="VAR",
            simple=False
        )
        self.assertIsNone(
            np.testing.assert_array_equal(
                test_predictions.columns.values, self.expected_pred_col_names
            )
        )

        ###################
        ## ML Dictionary ##
        ###################
        self.assertIsNone(
            np.testing.assert_array_equal(
                np.round(ml_dict.get('VAR').get('forecast')[0]['mean'].values.astype(np.double), 6),
                self.forecast_gold_var_multivar_internal_val_cv_fold1,

            ),
            "(Multivar Test) VAR Forecast does not match up with expected values."
        )
        self.assertIsNone(
            np.testing.assert_array_equal(
                np.round(ml_dict.get('VAR').get('forecast')[1]['mean'].values.astype(np.double), 6),
                self.forecast_gold_var_multivar_internal_val_cv_fold2
            ),
            "(Multivar Test) VAR Forecast does not match up with expected values."
        )

        self.assertEqual(
            round(ml_dict.get('VAR').get('rmse')[0], 8), self.rmse_gold_var_multivar_cv_fold1,
            "(Multivar Test) VAR RMSE does not match up with expected values.")
        self.assertEqual(
            round(ml_dict.get('VAR').get('rmse')[1], 8), self.rmse_gold_var_multivar_cv_fold2,
            "(Multivar Test) VAR RMSE does not match up with expected values.")


    def test_univar(self):
        """
        Test 3: Univariate VAR
        """
        print("\n\n" + "*"*50)
        print("Performing Unit Test: 'test_univar'")
        print("*"*50 + "\n\n")

        automl_model = ATS(
            score_type='rmse', forecast_period=self.forecast_period, time_interval='Month',
            non_seasonal_pdq=None, seasonality=False, seasonal_period=12,
            model_type='VAR',
            verbose=0)
        automl_model.fit(
            traindata=self.train_univar,
            ts_column=self.ts_column,
            target=self.target,
            cv=None
            )
        ml_dict = automl_model.get_ml_dict()

        self.assertIsNone(automl_model.get_model_build('VAR'), "Expected Univar VAR model to be None but did not get None.")

        # Simple forecast with forecast window = one used in training
        # Using named model
        test_predictions = automl_model.predict(
            forecast_period=self.forecast_period,
            model="VAR"
        )
        self.assertIsNone(test_predictions)

        # Simple forecast with forecast window != one used in training
        # Using named model
        test_predictions = automl_model.predict(
            forecast_period=10,
            model="VAR"
        )
        self.assertIsNone(test_predictions)

        # Complex forecasts (returns confidence intervals, etc.)
        test_predictions = automl_model.predict(
            forecast_period=self.forecast_period,
            model="VAR",
            simple=False
        )
        self.assertIsNone(test_predictions)

        ###################
        ## ML Dictionary ##
        ###################
        self.assertEqual(
            ml_dict.get('VAR').get('forecast'), self.forecast_gold_var_univar,
            "(Univar Test) VAR Forecast does not match up with expected values."
        )

        self.assertEqual(
            round(ml_dict.get('VAR').get('rmse'), 8), self.rmse_gold_var_univar,
            "(Univar Test) VAR RMSE does not match up with expected values.")


from itertools import cycle
import numpy as np # type: ignore
import pandas as pd # type: ignore
import matplotlib.dates as mdates # type: ignore
from itertools import cycle
import matplotlib.pyplot as plt # type: ignore
import seaborn as sns # type: ignore
# This gives an error when running from a python script.
# Maybe, this should be set in the jupyter notebook directly.
# get_ipython().magic('matplotlib inline')
sns.set(style="white", color_codes=True)
# TSA from Statsmodels
import statsmodels.tsa.api as smt # type: ignore
import pdb
import copy
import dask
import dask.dataframe as dd

def time_series_plot(y, lags=31, title='Original Time Series', chart_type='line',
                     chart_freq='years'):
    """
    Plot a Time Series along with how it will look after differencing and what its
    AR/MA lags will be by viewing the ACF and PACF, along with its histogram.
    You just need to provide the time series (y) as a Series. Index is assumed
    to be Pandas datetime. It assumes that you want to see default lags of 31.
    But you can modify it to suit.
    """
    
    y = copy.deepcopy(y)
    if chart_freq in ['MS', 'M', 'SM', 'BM', 'CBM', 'SMS', 'BMS']:
        chart_time = 'months'
    elif chart_freq in ['D', 'B', 'C']:
        chart_time = 'days'
    elif chart_freq in ['W']:
        chart_time = 'weeks'
    elif chart_freq in ['Q', 'BQ', 'QS', 'BQS']:
        chart_time = 'quarters'
    elif chart_freq in ['A,Y', 'BA,BY', 'AS,YS', 'BAS,YAS']:
        chart_time = 'years'
    elif chart_freq in ['BH', 'H', 'h']:
        chart_time = 'hours'
    elif chart_freq in ['T,min']:
        chart_time = 'minutes'
    elif chart_freq in ['S', 'L,milliseconds', 'U,microseconds', 'N,nanoseconds']:
        chart_time = 'seconds'
    else:
        print('chart frequency not known. Continuing...')
        return
    colors = cycle('byrcmgkbyrcmgkbyrcmgkbyrcmgkbyr')
    fig = plt.figure(figsize=(20, 20))
    grid = plt.GridSpec(3, 2, wspace=0.5, hspace=0.5)
    fig.subplots_adjust(hspace=1)
    ########## Use the gridspec function ##############
    ts_ax = plt.subplot(grid[0, 0:])
    diff_ax = plt.subplot(grid[1, 0])
    hist_ax = plt.subplot(grid[1, 1])
    acf_ax = plt.subplot(grid[2, 0])
    pacf_ax = plt.subplot(grid[2, 1])
    ### Draw multiple kinds of graphs here to each subplot axis ###

    if type(y) == dask.dataframe.core.DataFrame or type(y) == dask.dataframe.core.Series:
            y = y.head(len(y)) ## this converts it into a pandas Series
    if chart_type == 'line':
        y.plot(ax=ts_ax, color=next(colors))
    else:
        if chart_time == 'years':
            majors = mdates.YearLocator()  # every year
            minors = mdates.MonthLocator()  # every month
            majorsFmt = mdates.DateFormatter('%Y')
        elif chart_time == 'months':
            majors = mdates.YearLocator()  # every year
            minors = mdates.MonthLocator()  # every month
            majorsFmt = mdates.DateFormatter('\n\n\n%b\n%Y')
        elif chart_time == 'weeks':
            majors = mdates.MonthLocator()
            minors = mdates.WeekdayLocator(byweekday=(1), interval=1)
            majorsFmt = mdates.DateFormatter('\n\n\n%b\n%Y')
        elif chart_time == 'days':
            majors = mdates.DayLocator(bymonthday=None, interval=1, tz=None)
            minors = mdates.HourLocator(byhour=None, interval=1, tz=None)
            majorsFmt = mdates.DateFormatter('\n\n\n%d\n%b')
        else:
            majors = mdates.YearLocator()  # every year
            minors = mdates.MonthLocator()  # every month
            majorsFmt = mdates.DateFormatter('\n\n\n%b\n%Y')
        try:
            #### this works in most cases but in some cases, it gives an error
            ts_ax.bar(y.index, height=y, width=20, color=list((y>0).astype(int).map({1:'g',0:'r'}).values))
        except:
            #### In some cases where y is a dataframe, this might work.
            yindex = y.index
            yvalues = y.values.ravel()
            ts_ax.bar(yindex, height=yvalues, width=20, color=list(using_where((yvalues>0).astype(int)).ravel()))
        ts_ax.xaxis.set_major_locator(majors)
        ts_ax.xaxis.set_major_formatter(majorsFmt)
        ts_ax.xaxis.set_minor_locator(minors)
        ts_ax.format_xdata = mdates.DateFormatter('%Y-%m-%d')
        ts_ax.grid(True)
    #### Now draw the ACF and PACF charts
    ts_ax.set_title(title)
    y.diff(1).plot(ax=diff_ax, color=next(colors))
    diff_ax.set_title('After Differencing = 1')
    y.plot(ax=hist_ax, kind='hist', bins=25, color=next(colors))
    hist_ax.set_title('Histogram for Original Series')
    try:
        smt.graphics.plot_acf(y, lags=lags, ax=acf_ax)
        acf_ax.set_title('ACF for Original Series')
    except:
        acf_ax.set_title('Data Error: Could not draw ACF for Original Series')
    try:
        ### the number of lags cannot be greater than 50% of len of y. So limit it.
        if lags  >= len(y)*0.5:
            lags = int(len(y)*0.5 - 1)
        smt.graphics.plot_pacf(y, lags=lags, ax=pacf_ax)
        pacf_ax.set_title('PACF for Original Series')
    except:
        pacf_ax.set_title('Data Error: Could not draw PACF for Original Series')
    [ax.set_xlim(0) for ax in [acf_ax, pacf_ax]]
    plt.show(block=False)

def using_where(x):
    return np.where(x == 1, 'g', 'r')
#################################################################################

def top_correlation_to_name(stocks, column_name, searchstring, top=5):
    """
    ####################################################################################
    This function draws a correlation chart of the top "x" rows of a data frame that are highly
    correlated to a selected row in the dataframe. You can think of the rows of the input
    dataframe as containing stock prices or fund flows or product sales and the columns should
    contain time series data of prices or flows or sales over multiple time periods.
    Now this program will allow you to select the top 5 or 10 rows that are highly correlated
    to a given row selected by the column: column_name and using a search string "searchstring".
    The program will search for the search string in that column column_name and return a list
    of 5 or 10 rows that are the most correlated to that selected row. If you give "top" as
    a float ratio then it will use the ratio as the cut off point in the correlation
    coefficient to select rows.
    ####################################################################################
    """
    #### First increment top by 1 since you are asking for top X names in addition to the one you have, top += 1
    incl = [x for x in list(stocks) if x not in column_name]
    ### First drop all NA rows since they will mess up your correlations, stocks.dropna(inplace=True)
    if stocks.empty:
        print('After dropping NaNs, the data frame has become empty.')
        return
    ### Now find the highest correlated rows to the selected row ###
    try:
        index_val = search_string(stocks, column_name,searchstring).index[0]
    except:
        print('Not able to find the search string in the column.')
        return
    ### Bring that selected Row to the top of the Data Frame
    df = stocks[:]
    # TODO: Undefined variable 'l'
    df["new"] = range(l, len(df)+l)
    df.loc[index_val,"new"] = 0
    stocks = df.sort_values("new").drop("new",axis=1)
    stocks.reset_index(inplace=True,drop=True)
    ##### Now calculate the correlation coefficients of other rows with the Top row
    try:
        cordf = pd.DataFrame(stocks[incl].T.corr().sort_values(0, ascending=False))
    except:
        print('Cannot calculate Correlations since Dataframe contains string values or objects.')
        return
    try:
        cordf = stocks[column_name].join(cordf)
    except:
        cordf = pd.concat((stocks[column_name], cordf), axis=1)
    #### Visualizing the top 5 or 10 or whatever cut-off they have given for Corr Coeff
    if top >= 1:
        top10index = cordf.sort_values(0, ascending=False).iloc[:top, :3].index
        top10names = cordf.sort_values(0, ascending=False).iloc[:top, :3][column_name]
        top10values = cordf.sort_values(0, ascending=False)[0].values[:top]
    else:
        top10index = cordf.sort_values(0, ascending=False)[
                     cordf.sort_values(0, ascending=False)[0].values >= top].index
        top10names = cordf.sort_values(0, ascending=False)[cordf.sort_values(
                                       0, ascending=False)[0].values >= top][column_name]
        top10alues = cordf.sort_values(0, ascending=False)[cordf.sort_values(
                                       0, ascending=False)[0].values >= top][0]
    print(top10names, top10values)
    #### Now plot the top rows that are highly correlated based on condition above
    stocksloc = stocks.iloc[top10index]
    #### Visualizing using Matplotlib ###
    stocksloc = stocksloc.T
    stocksloc = stocksloc.reset_index(drop=True)
    stocksloc.columns = stocksloc.iloc[0].values.tolist()
    stocksloc.drop(0).plot(subplots=True, figsize=(15, 10), legend=False,
                           title="Top %s Correlations to %s" % (top, searchstring))
    [ax.legend(loc=1) for ax in plt.gcf().axes]
    plt.tight_layout()
    plt.show(block=False)

################################################################################
def pretty_print_table(dfo):
    from io import StringIO
    import prettytable
    output = StringIO()
    dfo.to_csv(output)
    output.seek(0)
    pt = prettytable.from_csv(output)
    print(pt)

import copy
def test_stationarity(timeseries, maxlag=31, regression='c', autolag='BIC',
                      window=None, plot=False, verbose=False):
    """
    Check unit root stationarity of a time series array or an entire dataframe.
    Note that you must send in a dataframe as df.values.ravel() - otherwise ERROR.
    Null hypothesis: the series is non-stationary.
    If p >= alpha, the series is non-stationary.
    If p < alpha, reject the null hypothesis (has unit root stationarity).
    Original source: http://www.analyticsvidhya.com/blog/2016/02/time-series-forecasting-codes-python/
    Function: http://statsmodels.sourceforge.net/devel/generated/statsmodels.tsa.stattools.adfuller.html
    window argument is only required for plotting rolling functions. Default=4.
    """
    timeseries = copy.deepcopy(timeseries)
    if len(timeseries) <= int(1.5*maxlag):
        maxlag = 5  ## set it to a low number
    # set defaults (from function page)
    if type(timeseries) == pd.DataFrame:
        print('modifying time series dataframe into an array to test')
        timeseries = timeseries.values.ravel()
    if regression is None:
        regression = 'c'
    if verbose:
        print('Running Augmented Dickey-Fuller test with paramters:')
        print('    maxlag: {}'.format(maxlag),'regression: {}'.format(regression),'autolag: {}'.format(autolag))
    alpha = 0.05
    if plot:
        try:
            if window is None:
                window = 4
            # Determing rolling statistics
            rolmean = timeseries.rolling(window=window, center=False).mean()
            rolstd = timeseries.rolling(window=window, center=False).std()
            # Plot rolling statistics:
            orig = plt.plot(timeseries, color='blue', label='Original')
            mean = plt.plot(rolmean, color='red', label='Rolling Mean ({})'.format(window))
            std = plt.plot(rolstd, color='black', label='Rolling Std ({})'.format(window))
            plt.legend(loc='best')
            plt.title('Rolling Mean & Standard Deviation')
            plt.show(block=False)
        except:
            print('Data must have date-time as index to plot!')
            return
    # Perform Augmented Dickey-Fuller test:
    try:
        ### Use Statsmodels for tests ###########
        dftest = smt.adfuller(timeseries, maxlag=maxlag, regression=regression, autolag=autolag)
        dfoutput = pd.Series(dftest[0:4], index=['Test Statistic',
                                                 'p-value',
                                                 '#Lags Used',
                                                 'Number of Observations Used',
                                                 ],name='Dickey-Fuller Augmented Test')
        for key, value in dftest[4].items():
            dfoutput['Critical Value (%s)' % key] = value
        if verbose:
            print('Results of Augmented Dickey-Fuller Test:')
            pretty_print_table(dfoutput)
        if dftest[1] >= alpha:
            print(' this series is non-stationary. Trying test again after differencing...')
            timeseries = pd.Series(timeseries).diff(1).dropna().values
            dftest = smt.adfuller(timeseries, maxlag=maxlag, regression=regression, autolag=autolag)
            dfoutput = pd.Series(dftest[0:4], index=['Test Statistic',
                                                     'p-value',
                                                     '#Lags Used',
                                                     'Number of Observations Used',
                                                     ],name='Dickey-Fuller Augmented Test')
            for key, value in dftest[4].items():
                dfoutput['Critical Value (%s)' % key] = value
            if verbose:
                print('After differencing=1, results of Augmented Dickey-Fuller Test:')
                pretty_print_table(dfoutput)
            if dftest[1] >= alpha:
                print(' this series is not stationary')
                return False
            else:
                print(' this series is stationary')
                return True
        else:
            print(' this series is stationary')
            return True
    except:
        print('Error: Stationary test failed. Data must be np.array. Check your input and try stationary test again')
        return
################################################################################



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="auto_ts",
    version="0.0.46",
    author="Ram Seshadri",
    # author_email="author@example.com",
    description="Automatically Build Multiple Time Series models fast - now with Facebook Prophet!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='Apache License 2.0',
    url="https://github.com/AutoViML/Auto_TS",
    packages=setuptools.find_packages(exclude=("tests",)),
    install_requires=[
        "ipython",
        "jupyter",
        "pmdarima",
        "numpy",
        "xlrd",
        "pandas",
        "matplotlib",
        "seaborn",
        "scikit-learn>=0.24.0",
        "fbprophet",
        "statsmodels",
        "xgboost==0.90",
        "prettytable",
        "dask>=2.30.0",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



"""Module to build a VAR model
"""
import pdb
from typing import Optional
import warnings
import itertools
import operator
import copy

import numpy as np  # type: ignore
import pandas as pd # type: ignore
from pandas.core.generic import NDFrame # type:ignore
import dask
import dask.dataframe as dd

import matplotlib.pyplot as plt # type: ignore
import seaborn as sns # type: ignore
sns.set(style="white", color_codes=True)

from statsmodels.tsa.statespace.varmax import VARMAX # type: ignore

#from tscv import GapWalkForward # type: ignore
from sklearn.model_selection import TimeSeriesSplit

# helper functions
from ...utils import print_dynamic_rmse
from ...models.ar_based.param_finder import find_lowest_pq
from ..build_base import BuildBase


class BuildVAR(BuildBase):
    """Class to build a VAR model
    """
    def __init__(self, scoring, forecast_period=2, p_max=3, q_max=3, verbose=0):
        """
        Automatically build a VAR Model

        Since it automatically builds a VAR model, you need to give it a Criteria (scoring) to optimize
        on. You can give it any of the following metrics as scoring options:
            AIC, BIC, Deviance, Log-likelihood.
        You can give the highest order values for p and q. Default is set to 3 for both.
        """
        super().__init__(
            scoring=scoring,
            forecast_period=forecast_period,
            verbose=verbose
        )
        self.p_max = p_max
        self.q_max = q_max
        self.best_p = None
        self.best_d = None
        self.best_q = None

    # def fit(self, ts_df):
    def fit(self, ts_df: pd.DataFrame, target_col: str, cv: Optional[int] = None) -> object:
        """
         This builds a VAR model given a multivariate time series data frame with time as the Index.

        :param ts_df The time series data to be used for fitting the model. Note that the input can be
        a data frame with one column or multiple cols or a multivariate array. However, the first column
        must be the target variable. You must include only Time Series data in it. DO NOT include
        "Non-Stationary" or "Trendy" data. Make sure your Time Series is "Stationary" before you send
        it in!! If not, this will give spurious results.
        :type ts_df pd.DataFrame

        :param target_col The column name of the target time series that needs to be modeled.
        All other columns will be considered as exogenous variables (if applicable to method)
        :type target_col str

        :param cv: Number of folds to use for cross validation.
        Number of observations in the Validation set for each fold = forecast period
        If None, a single fold is used
        :type cv Optional[int]

        :rtype object
        """
        self.original_target_col = target_col
        self.original_preds = [x for x in list(ts_df) if x not in [self.original_target_col]]

        ts_df = ts_df[[self.original_target_col] + self.original_preds]

        #######################################
        #### Cross Validation across Folds ####
        #######################################

        rmse_folds = []
        norm_rmse_folds = []
        forecast_df_folds = []
        norm_rmse_folds2 = []

        ### Creating a new way to skip cross validation when trying to run auto-ts multiple times. ###
        if cv == 0:
            cv_in = 0
        else:
            cv_in = copy.deepcopy(cv)
        NFOLDS = self.get_num_folds_from_cv(cv)
        #cv = GapWalkForward(n_splits=NFOLDS, gap_size=0, test_size=self.forecast_period)
        #cv = TimeSeriesSplit(n_splits=NFOLDS, test_size=self.forecast_period) ### sklearn version 0.0.24
        max_trainsize = len(ts_df) - self.forecast_period
        try:
            cv = TimeSeriesSplit(n_splits=NFOLDS, test_size=self.forecast_period) ### this works only sklearn v 0.0.24]
        except:
            cv = TimeSeriesSplit(n_splits=NFOLDS, max_train_size = max_trainsize)

        if type(ts_df) == dask.dataframe.core.DataFrame:
            ts_df = dft.head(len(ts_df)) ### this converts dask into a pandas dataframe

        if  cv_in == 0:
            print('Skipping cross validation steps since cross_validation = %s' %cv_in)
            self.find_best_parameters(data = ts_df)
            y_train = ts_df.iloc[:, [0, self.best_d]]
            bestmodel = self.get_best_model(y_train)
            self.model = bestmodel.fit(disp=False)
        else:
            for fold_number, (train_index, test_index) in enumerate(cv.split(ts_df)):
                dftx = ts_df.head(len(train_index)+len(test_index))
                ts_train = dftx.head(len(train_index)) ## now train will be the first segment of dftx
                ts_test = dftx.tail(len(test_index)) ### now test will be right after train in dftx

                print(f"\nFold Number: {fold_number+1} --> Train Shape: {ts_train.shape[0]} Test Shape: {ts_test.shape[0]}")
                self.find_best_parameters(data = ts_train)

                #########################################
                #### Define the model with fold data ####
                #########################################
                y_train = ts_train.iloc[:, [0, self.best_d]]
                bestmodel = self.get_best_model(y_train)

                ######################################
                #### Fit the model with fold data ####
                ######################################

                if self.verbose >= 1:
                    print(f'Fitting best VAR model on Fold: {fold_number+1}')
                try:
                    self.model = bestmodel.fit(disp=False)
                except Exception as e:
                    print(e)
                    print(f'Error: VAR Fit on Fold: {fold_number+1} unsuccessful.')
                    return bestmodel, None, np.inf, np.inf

                forecast_df = self.predict(ts_test.shape[0],simple=False)
                forecast_df_folds.append(forecast_df['yhat'].values)

                rmse, norm_rmse = print_dynamic_rmse(ts_test.iloc[:, 0].values, forecast_df['yhat'].values,
                                            ts_train.iloc[:, 0].values)
                rmse_folds.append(rmse)
                norm_rmse_folds.append(norm_rmse)

            norm_rmse_folds2 = rmse_folds/ts_df[self.original_target_col].values.std()  # Same as what was there in print_dynamic_rmse()
            self.model.plot_diagnostics(figsize=(16, 12))
            axis = self.model.impulse_responses(12, orthogonalized=True).plot(figsize=(12, 4))
            axis.set(xlabel='Time Steps', title='VAR model Impulse Response Functions')

        ###############################################
        #### Refit the model on the entire dataset ####
        ###############################################
        y_train = ts_df.iloc[:, [0, self.best_d]]
        self.refit(ts_df=y_train)

        # return self.model, forecast_df_folds, rmse_folds, norm_rmse_folds
        return self.model, forecast_df_folds, rmse_folds, norm_rmse_folds2

    def predict(
            self,
            testdata: Optional[pd.DataFrame] = None,
            forecast_period: Optional[int] = None,
            simple: bool = True
        ) -> NDFrame:
        """
        Return the predictions
        """

        if testdata is not None:
            warnings.warn(
                "You have passed exogenous variables to make predictions for a VAR model. " +
                "VAR model will predict all exogenous variables automatically, " +
                "hence your passed values will not be used."
            )
            if isinstance(testdata, pd.DataFrame) or isinstance(testdata, pd.Series):
                if len(testdata) != self.forecast_period:
                    self.forecast_period = testdata.shape[0]
            elif isinstance(testdata, int):
                self.forecast_period = testdata

            forecast_period = self.forecast_period

        # Extract the dynamic predicted and true values of our time series
        if forecast_period is None:
            # use the forecast period used during training
            forecast_period = self.forecast_period

        # y_forecasted = self.model.forecast(forecast_period)

        res = self.model.get_forecast(forecast_period)
        res_frame = res.summary_frame()

        res_frame.rename(columns={'mean':'yhat'},inplace=True)

        if simple:
            res_frame = res_frame['yhat']
            res_frame = res_frame.squeeze() # Convert to a pandas series object
        else:
            # Pass as is
            pass

        return res_frame


    def find_best_parameters(self, data: pd.DataFrame):
        """
        Given a dataset, finds the best parameters using the settings in the class
        """
        #### dmax here means the column number of the data frame: it serves as a placeholder for columns
        dmax = data.shape[1]
        ###############################################################################################
        cols = data.columns.tolist()
        # TODO: #14 Make sure that we have a way to not rely on column order to determine the target
        # It is assumed that the first column of the dataframe is the target variable ####
        ### make sure that is the case before doing this program ####################
        i = 1
        results_dict = {}

        for d_val in range(1, dmax):
            # Takes the target column and one other endogenous column at a time
            # and makes a prediction based on that. Then selects the best
            # exogenous column at the end.
            y_train = data.iloc[:, [0, d_val]]
            print('\nAdditional Variable in VAR model = %s' % cols[d_val])
            info_criteria = pd.DataFrame(
                index=['AR{}'.format(i) for i in range(0, self.p_max+1)],
                columns=['MA{}'.format(i) for i in range(0, self.q_max+1)]
            )
            for p_val, q_val in itertools.product(range(0, self.p_max+1), range(0, self.q_max+1)):
                if p_val == 0 and q_val == 0:
                    info_criteria.loc['AR{}'.format(p_val), 'MA{}'.format(q_val)] = np.nan
                    print(' Iteration %d completed' % i)
                    i += 1
                else:
                    try:
                        model = VARMAX(y_train, order=(p_val, q_val), trend='c')
                        model = model.fit(max_iter=1000, disp=False)
                        info_criteria.loc['AR{}'.format(p_val), 'MA{}'.format(q_val)] = eval('model.' + self.scoring)
                        print(' Iteration %d completed' % i)
                        i += 1
                    except Exception:
                        i += 1
                        print(' Iteration %d completed' % i)
            info_criteria = info_criteria[info_criteria.columns].astype(float)
            interim_d = copy.deepcopy(d_val)
            interim_p, interim_q, interim_bic = find_lowest_pq(info_criteria)
            if self.verbose == 1:
                _, axis = plt.subplots(figsize=(20, 10))
                axis = sns.heatmap(
                    info_criteria,
                    mask=info_criteria.isnull(),
                    ax=axis,
                    annot=True,
                    fmt='.0f'
                )
                axis.set_title(self.scoring)
            results_dict[str(interim_p) + ' ' + str(interim_d) + ' ' + str(interim_q)] = interim_bic
        best_bic = min(results_dict.items(), key=operator.itemgetter(1))[1]
        best_pdq = min(results_dict.items(), key=operator.itemgetter(1))[0]
        self.best_p = int(best_pdq.split(' ')[0])
        self.best_d = int(best_pdq.split(' ')[1])
        self.best_q = int(best_pdq.split(' ')[2])
        print('Best variable selected for VAR: %s' % data.columns.tolist()[self.best_d])

    def refit(self, ts_df: pd.DataFrame) -> object:
        """
        Refits an already trained model using a new dataset
        Useful when fitting to the full data after testing with cross validation
        :param ts_df The time series data to be used for fitting the model
        :type ts_df pd.DataFrame
        :rtype object
        """
        bestmodel = self.get_best_model(ts_df)
        print('Refitting data with previously found best parameters')
        try:
            self.model = bestmodel.fit(disp=False)
            print('    Best %s metric = %0.1f' % (self.scoring, eval('self.model.' + self.scoring)))
        except Exception as exception:
            print(exception)

        return self


    def get_best_model(self, data: pd.DataFrame):
        """
        Returns the 'unfit' SARIMAX model with the given dataset and the
        selected best parameters. This can be used to fit or refit the model.
        """
        bestmodel = VARMAX(data, order=(self.best_p, self.best_q), trend='c')
        return bestmodel
